
//		APing.cpp  --  Adds an audible ping command to TCC.


#define UNICODE						1
#define _UNICODE					1
#define _CRT_RAND_S

#include <stdio.h>
#include <tchar.h>
#include <WS2tcpip.h>
#include <windows.h>
#include <IPHlpApi.h>
#include <WinDNS.h>
#include <IcmpAPI.h>
#include <CommCtrl.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					1
#define VER_MINOR					0
#define VER_BUILD					3
#define VER_PATCH					1

#define PLUGIN_NAME					L"APing"

#define MAX_ARG_LEN					8192
#define SHORT_BUF_LEN				1024
#define IP_STRING_LEN				128


#define PING_MIN_TIMEOUT			1
#define PING_MAX_TIMEOUT			60000
#define PING_DEFAULT_TIMEOUT		3000
#define PING_MIN_PKTSIZE			12
#define PING_MAX_PKTSIZE			65520
#define PING_DEFAULT_PKTSIZE		64
#define TTL_MIN						1
#define TTL_MAX						255
#define TTL_DEFAULT					128
#define MIN_SNOOZE_TIME				250
#define MAX_SNOOZE_TIME				10000
#define DEFAULT_SNOOZE_TIME			1000
#define DEFAULT_ERROR_SNOOZE		650
#define MIN_INSTRUMENT				1
#define MAX_INSTRUMENT				128
#define DEFAULT_INSTRUMENT			115
#define MIN_DRUM_KEY				35
#define MAX_DRUM_KEY				81
#define DEFAULT_ERROR_DRUM			66
#define OP_ERROR_DRUM				58

#define FLAG_CONTROL_C				1
#define FLAG_CONTROL_BREAK			2

#define PING_FLAG_TRY_V6_FIRST		1
#define PING_FLAG_COMPARE_DATA		2
#define PING_FLAG_DEBUG				128

#define ERR_DATA_MISMATCH			-10
#define ERR_UNKNOWN					-20

#define LOCALHOSTADDR				0x0100007f

#define MAX_FILENAME_LEN			32768

#define ERRHANDLE GetStdHandle( STD_ERROR_HANDLE )

#define CR							0x0d
#define LF							0x0a



#ifdef _WIN64
#define PLATFORM					L"x64"
#define MIDI_HANDLE_VAR_TYPE		REG_QWORD
#else
#define PLATFORM					L"x86"
#define MIDI_HANDLE_VAR_TYPE		REG_DWORD
#endif


struct __IO_STATUS_BLOCK {
	union {
		NTSTATUS Status;
		PVOID Pointer;
	};
	ULONG_PTR Information;
};




//		Global variables:

WSADATA WsaData;

HMIDIOUT MidiHandle					= NULL;
BOOL Silence						= FALSE;

unsigned int ConsoleCtrlFlags		= 0;
unsigned int PacketsReceived		= 0;
unsigned int PacketsLost			= 0;
unsigned int MinPingTime			= UINT_MAX;
unsigned int MaxPingTime			= 0;
unsigned int TotalPingTime			= 0;


OSVERSIONINFO OSVerInfo;

HINSTANCE MyDLLHandle = NULL;



const wchar_t PluginFeatures[] = L"aping,apinghelp,apingmacro,@apingmacro,@lookuphost,@pingdhcp,\
@pingdns,@pingrouter,@pingtime,@pingwins,_pingavg,_pingmax,_pingmin,_pingslost,_pingsrcvd";




const wchar_t APingHelpText[] = L"Audibly ping a network host.\n\
\n\
APING /4 /6 /A /C /D /F /I:n /L:n /N:n /Q /T /W:n hostname  \r\n\
\n\
\t/4\tuse IPv4\n\
\t/6\tuse IPv6\n\
\t/A\tresolve numeric addresses\n\
\t/C\tshow line numbers\n\
\t/D\tdisable highlight\n\
\t/F\tset the Don\u2019t Fragment flag\n\
\t/I:\tspecify Time-To-Live\n\
\t/L:\tpacket size, in bytes    (64)\n\
\t/N:\tnumber of pings to send  (4)\n\
\t/Q\tquietly\n\
\t/T\tcontinue until Control-C hit\n\
\t/W:\ttimeout in milliseconds  (3000)\n\
\n\
Options may begin with either a forward slash or a dash.  \n\
The hostname must be specified.\n";


const wchar_t APingMacroHelpText[] = L"List or change APING macros.\n\
\n\
APINGMACRO /D /L /P #n=value...  \r\n\
\n\
\t/D\tdisable highlight\n\
\t/L\tlist all defined macros\n\
\t/P\tpage output\n\
\n\
Use #n=value to create a macro, #n= (with no value) to delete it.  \n\
With no arguments, the default is to list all.\n";


const wchar_t APingHelpHelpText[] = L"Open the APing plugin help file.\n\
\n\
APINGHELP /C /F /S /S:text /V topic\r\n\
\n\
\t/C\tselect the Contents tab\n\
\t/F\tselect the Favorites tab\n\
\t/S\tselect the Search tab\n\
\t/S:\tselect the Search tab and search for text\n\
\t/V\tshow detailed plugin version info\n\
\n\
Opens this plugin\u2019s help file to the specified topic.\n\
If no topic is given, opens the Overview page by default.  \n";



const wchar_t RegKeyName[]			= L"Software\\JPPlugins\\APing";
const wchar_t CountRegName[]		= L"Count";
const wchar_t TimeoutRegName[]		= L"Timeout";
const wchar_t PacketSizeRegName[]	= L"PacketSize";
const wchar_t DefaultHostRegName[]	= L"DefaultHostname";
const wchar_t InstrumentRegName[]	= L"Instrument";
const wchar_t OctaveRegName[]		= L"Octave";
const wchar_t ErrorDrumRegName[]	= L"ErrorDrum";
const wchar_t OpErrorDrumRegName[]	= L"OpErrorDrum";
const wchar_t SnoozeTimeRegName[]	= L"SnoozeTime";
const wchar_t ErrorSnoozeRegName[]	= L"ErrorSnooze";
const wchar_t NoMidiRegName[]		= L"NoMidi";

const wchar_t SharedMidiRegKey[]			= L"Software\\JPPlugins\\Shared MIDI\\%04X";
const wchar_t SharedMidiHandleValue[]		= L"Handle";
const wchar_t SharedMidiRefCountValue[]		= L"RefCount";


#define NUM_MAGIC_NAMES				5
#define MAGIC_ROUTER_NAME			0
#define MAGIC_ROUTER_NAME_2			1
#define MAGIC_DNS_NAME				2
#define MAGIC_DHCP_NAME				3
#define MAGIC_WINS_NAME				4

#define MAX_MACRO_NAME_LEN			10

const wchar_t MagicName[NUM_MAGIC_NAMES][8]		= { L"#router", L"#gw", L"#dns", L"#dhcp", L"#wins" };

const wchar_t RegMacroPrefix[]		= L"Host-";



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle( STD_ERROR_HANDLE ), shortbuf );

	return 0;
}


#include "ParseArgs.cpp"
#include "SoundVol.cpp"



DWORD RegGetDWord( LPCTSTR KeyName, LPCTSTR ValueName, DWORD Min, DWORD Max, DWORD Default )
{
	BOOL Good			= FALSE;
	HKEY KeyHandle		= NULL;
	DWORD Value			= 0;
	DWORD ValueType		= 0;
	DWORD ValueSize		= 0;


	if ( RegOpenKeyEx( HKEY_CURRENT_USER, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
		
		ValueType = 0;
		ValueSize = sizeof( DWORD );

		if ( RegQueryValueEx( KeyHandle, ValueName, NULL, &ValueType, (BYTE*) &Value, &ValueSize ) == ERROR_SUCCESS )
			if ( ( ValueType == REG_DWORD ) && ( Value >= Min ) && ( Value <= Max ) )
				Good = TRUE;

		RegCloseKey( KeyHandle );
	}


	if ( !Good ) {
		if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

			ValueType = 0;
			ValueSize = sizeof( DWORD );

			if ( RegQueryValueEx( KeyHandle, ValueName, NULL, &ValueType, (BYTE*) &Value, &ValueSize ) == ERROR_SUCCESS )
				if ( ( ValueType == REG_DWORD ) && ( Value >= Min ) && ( Value <= Max ) )
					Good = TRUE;

			RegCloseKey( KeyHandle );
		}
	}


	if ( Good )
		return Value;
	else
		return Default;
}



int RegGetString( LPCTSTR KeyName, LPCTSTR ValueName, LPTSTR outString )
{
	BOOL Good						= FALSE;
	HKEY KeyHandle					= NULL;
	DWORD ValueType					= 0;
	DWORD ValueSize					= 0;
	wchar_t temp[NI_MAXHOST]		= L"";


	if ( RegOpenKeyEx( HKEY_CURRENT_USER, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

		ValueType = 0;
		ValueSize = NI_MAXHOST * sizeof( wchar_t );
		memset( temp, 0x00, ValueSize );

		if ( RegQueryValueEx( KeyHandle, ValueName, NULL, &ValueType, (BYTE*) temp, &ValueSize ) == ERROR_SUCCESS )
			if ( ( ValueType == REG_SZ ) && temp[0] )
				Good = TRUE;

		RegCloseKey( KeyHandle );
	}


	if ( !Good ) {
		if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

			ValueType = 0;
			ValueSize = NI_MAXHOST * sizeof( wchar_t );
			memset( temp, 0x00, ValueSize );

			if ( RegQueryValueEx( KeyHandle, ValueName, NULL, &ValueType, (BYTE*) temp, &ValueSize ) == ERROR_SUCCESS )
				if ( ( ValueType == REG_SZ ) && temp[0] )
					Good = TRUE;

			RegCloseKey( KeyHandle );
		}
	}
	

	if ( Good ) {
		wcscpy_s( outString, NI_MAXHOST, temp );
		return (int) wcslen( temp );
	}
	else
		return 0;
}



BOOL CheckMidiDisabled()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";
	DWORD Disabled						= 0;


	GetEnvironmentVariable( NoMidiRegName, temp, SHORT_BUF_LEN );
	if ( temp[0] )
		return TRUE;

	Disabled = RegGetDWord( RegKeyName, NoMidiRegName, 0, -1, 0 );

	if ( Disabled )
		return TRUE;
	else
		return FALSE;
}



HMIDIOUT OpenSharedMidi()
{
	wchar_t KeyName[SHORT_BUF_LEN]		= L"";

	HKEY KeyHandle			= NULL;
	HMIDIOUT MidiHandle		= NULL;
	DWORD RefCount			= 0;
	DWORD VarType			= 0;
	DWORD VarSize			= 0;



	wsprintf( KeyName, SharedMidiRegKey, GetCurrentProcessId() );

	RegCreateKeyEx( HKEY_CURRENT_USER, KeyName, 0, NULL, REG_OPTION_VOLATILE, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL );


	if ( KeyHandle ) {
		VarType = 0;
		VarSize = sizeof( HMIDIOUT );
		RegQueryValueEx( KeyHandle, SharedMidiHandleValue, NULL, &VarType, (BYTE*) &MidiHandle, &VarSize );

		if ( !MidiHandle || ( VarType != MIDI_HANDLE_VAR_TYPE ) )
			MidiHandle = NULL;

		if ( MidiHandle ) {
			VarType = 0;
			VarSize = sizeof( DWORD );
			RegQueryValueEx( KeyHandle, SharedMidiRefCountValue, NULL, &VarType, (BYTE*) &RefCount, &VarSize );

			if ( !RefCount || ( VarType != REG_DWORD ) )
				RefCount = 0, MidiHandle = NULL;
		}
	}


	if ( MidiHandle )
		RefCount++;
	else {
		midiOutOpen( &MidiHandle, 0, NULL, NULL, CALLBACK_NULL );

		if ( MidiHandle )
			RefCount = 1;
		else
			RefCount = 0;
	}


	if ( MidiHandle ) {
		if ( KeyHandle ) {
			RegSetValueEx( KeyHandle, SharedMidiHandleValue, 0, MIDI_HANDLE_VAR_TYPE, (BYTE*) &MidiHandle, sizeof( HMIDIOUT ) );
			RegSetValueEx( KeyHandle, SharedMidiRefCountValue, 0, REG_DWORD, (BYTE*) &RefCount, sizeof( DWORD ) );
			RegCloseKey( KeyHandle );
		}
	}
	else {
		RegCloseKey( KeyHandle );
		RegDeleteKey( HKEY_CURRENT_USER, KeyName );
	}


	return MidiHandle;
}



int CloseSharedMidi( HMIDIOUT MidiHandle )
{
	if ( !MidiHandle )
		return 0;


	wchar_t KeyName[SHORT_BUF_LEN]		= L"";

	HKEY KeyHandle			= NULL;
	HMIDIOUT RegMidiHandle	= NULL;
	DWORD RefCount			= 1;
	DWORD VarType			= 0;
	DWORD VarSize			= 0;


	wsprintf( KeyName, SharedMidiRegKey, GetCurrentProcessId() );

	RegCreateKeyEx( HKEY_CURRENT_USER, KeyName, 0, NULL, REG_OPTION_VOLATILE, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL );

	
	if ( KeyHandle ) {
		VarType = 0;
		VarSize = sizeof( HMIDIOUT );
		RegQueryValueEx( KeyHandle, SharedMidiHandleValue, NULL, &VarType, (BYTE*) &RegMidiHandle, &VarSize );

		if ( !RegMidiHandle || ( VarType != MIDI_HANDLE_VAR_TYPE ) )
			RegMidiHandle = NULL;


		if ( RegMidiHandle != MidiHandle ) {
			RegCloseKey( KeyHandle );
			midiOutClose( MidiHandle );
			return 0;
		}


		VarType = 0;
		VarSize = sizeof( DWORD );
		RegQueryValueEx( KeyHandle, SharedMidiRefCountValue, NULL, &VarType, (BYTE*) &RefCount, &VarSize );

		if ( ( RefCount < 1 ) || ( VarType != REG_DWORD ) )
			RefCount = 1;
	}


	RefCount--;


	if ( RefCount ) {
		if ( KeyHandle ) {
			RegSetValueEx( KeyHandle, SharedMidiRefCountValue, 0, REG_DWORD, (BYTE*) &RefCount, sizeof( DWORD ) );
			RegCloseKey( KeyHandle );
		}
	}
	else {
		midiOutClose( MidiHandle );
		RegCloseKey( KeyHandle );
		RegDeleteKey( HKEY_CURRENT_USER, KeyName );
	}

	return 0;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



#include "conlist.cpp"
#include "NewHelp.cpp"



unsigned int DigitsIn( unsigned int n )
{
	int k = 1;


	while ( n >= 10 ) {
		n /= 10;
		k ++;
	}
	
	return k;
}



int ShowSpaces( int n )
{
	if ( n < 1 )
		return 0;
	else if ( n >= SHORT_BUF_LEN )
		return 0;


	wchar_t temp[SHORT_BUF_LEN]			= L"";


	for ( int i = 0; i < n; i++ )
		temp[i] = 0x20;

	temp[n] = '\0';
	QPuts( temp );
	return n;
}



int SetMacroValue( unsigned int Number, LPTSTR Value )
{
	wchar_t RegValueName[SHORT_BUF_LEN]			= L"";
	HKEY KeyHandle								= NULL;

	BOOL Delete		= TRUE;
	int rv			= 1;
	

	if ( Value )
		if ( Value[0] )
			Delete = FALSE;

	wsprintf( RegValueName, L"%s%u", RegMacroPrefix, Number );

	
	if ( RegCreateKeyEx( HKEY_CURRENT_USER, RegKeyName, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_SET_VALUE,  NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {

		if ( Delete ) {
			if ( RegDeleteValue( KeyHandle, RegValueName ) == ERROR_SUCCESS )
				rv = 0;
		}
		else {
			if ( RegSetValueEx( KeyHandle, RegValueName, 0, REG_SZ, (BYTE*) Value, sizeof( wchar_t ) * ( (unsigned int) wcslen( Value ) + 1 ) ) == ERROR_SUCCESS )
				rv = 0;
		}

		RegCloseKey( KeyHandle );
	}
	
	return rv;
}



int SetMacroValue( LPTSTR Name, LPTSTR Value )
{
	if ( !Name )
		return -1;
	else if ( Name[0] == '\0' )
		return -1;


	wchar_t RegValueName[SHORT_BUF_LEN]			= L"";
	HKEY KeyHandle								= NULL;

	BOOL Delete		= TRUE;
	int rv			= 1;


	if ( Value )
		if ( Value[0] )
			Delete = FALSE;


	wcscpy_s( RegValueName, SHORT_BUF_LEN, RegMacroPrefix );
	wcscat_s( RegValueName, SHORT_BUF_LEN, Name );
	_wcsupr_s( RegValueName + wcslen( RegMacroPrefix ), SHORT_BUF_LEN );


	if ( RegCreateKeyEx( HKEY_CURRENT_USER, RegKeyName, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_SET_VALUE, NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {

		if ( Delete ) {
			if ( RegDeleteValue( KeyHandle, RegValueName ) == ERROR_SUCCESS )
				rv = 0;
		}
		else {
			if ( RegSetValueEx( KeyHandle, RegValueName, 0, REG_SZ, (BYTE*) Value, sizeof( wchar_t ) * ( (unsigned int) wcslen( Value ) + 1 ) ) == ERROR_SUCCESS )
				rv = 0;
		}

		RegCloseKey( KeyHandle );
	}

	return rv;
}



struct _ListMatchEntry {
	_ListMatchEntry *Next;
	unsigned int Number;
	wchar_t MacroName[1];
};

struct _SortMacroEntry {
	unsigned int Number;
	wchar_t MacroName[MAX_MACRO_NAME_LEN + 1];
	wchar_t Value[NI_MAXHOST];
};



int __cdecl MacroCmp$( void const *m1, void const *m2 )
{
	_SortMacroEntry *macro1 = (_SortMacroEntry*) m1, *macro2 = (_SortMacroEntry*) m2;
	int c					= wcscmp( macro1->MacroName, macro2->MacroName );


	if ( c )
		return c;

	if ( macro1->Number < macro2->Number )
		return -1;
	else if ( macro1->Number > macro2->Number )
		return 1;
	else
		return 0;
}



unsigned int IsMagicName( LPTSTR HostName, LPCTSTR Prefix )
{
	if ( !HostName || !Prefix )
		return 0;

	if ( ( HostName[0] == '\0' ) || ( Prefix[0] == '\0' ) )
		return 0;


	size_t PrefixLen		= wcslen( Prefix );


	if ( !_wcsicmp( HostName, Prefix ) )
		return 1;

	if ( !_wcsnicmp( HostName, Prefix, PrefixLen ) && ( HostName[PrefixLen] >= '1' ) && ( HostName[PrefixLen] <= '9' ) && ( HostName[PrefixLen + 1] == '\0' ) )
		return HostName[PrefixLen] - '0';

	return 0;
}



BOOL IsValidUserMacroName( LPTSTR Name )
{
	if ( !Name )
		return FALSE;
	else if ( !iswalpha( Name[0] ) )
		return FALSE;


	for ( int i = 1; Name[i]; i++ ) {

		if ( Name[i] == '\0' )
			break;

		if ( i >= MAX_MACRO_NAME_LEN )
			return FALSE;

		if ( !iswalnum( Name[i] ) && ( Name[i] != '_' ) )
			return FALSE;
	}

	for ( int i = 0; i < NUM_MAGIC_NAMES; i++ )
		if ( IsMagicName( Name, MagicName[i] + 1 ) )
			return FALSE;

	return TRUE;
}



BOOL MacroIsInList( LPTSTR MacroName, _ListMatchEntry *First )
{
	if ( !MacroName || !First )
		return FALSE;
	else if ( MacroName[0] == '\0' )
		return FALSE;


	_ListMatchEntry *This			= First;
	unsigned int Value				= 0;
	wchar_t *sep1					= NULL;
	BOOL Numeric					= FALSE;


	if ( iswdigit( MacroName[0] ) )
		if ( !ParseInt( MacroName, &Value, &sep1 ) )
			if ( *sep1 == '\0' )
				Numeric = TRUE;


	while ( This ) {

		if ( Numeric && ( This->MacroName[0] == '\0' ) ) {
			if ( This->Number == Value )
				return TRUE;
		}
		else if ( !Numeric && This->MacroName[0] ) {
			if ( !WildcardComparison( This->MacroName, MacroName, 0, 1 ) )
				return TRUE;
		}

		This = This->Next;
	}

	return FALSE;
}



int ListMacros( BOOL ShowCount, _ListMatchEntry *MatchEntries )
{
	HKEY KeyHandle						= NULL;
	wchar_t Name[SHORT_BUF_LEN]			= L"";
	wchar_t Value[NI_MAXHOST]			= L"";
	DWORD NameLen						= 0;
	DWORD ValueType						= 0;
	DWORD ValueSize						= 0;
	DWORD NumValues						= 0;

	unsigned int Number					= 0;
	wchar_t *sep1						= NULL;

	BOOL Match							= FALSE;

	_SortMacroEntry *MacroEntry			= NULL;

	unsigned int LongestName			= 0;

	int Count							= 0;



	CL_HookBreak();

	if ( RegOpenKeyEx( HKEY_CURRENT_USER, RegKeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

		RegQueryInfoKey( KeyHandle, NULL, NULL, NULL, NULL, NULL, NULL, &NumValues, NULL, NULL, NULL, NULL );

		if ( NumValues ) {

			MacroEntry = (_SortMacroEntry*) malloc( NumValues * sizeof( _SortMacroEntry ) );

			if ( MacroEntry ) {
				Count = 0;

				for ( DWORD Index = 0; TRUE; Index++ ) {

					NameLen		= SHORT_BUF_LEN;
					ValueType	= 0;
					ValueSize	= NI_MAXHOST * sizeof( wchar_t );
					Name[0]		= '\0';
					Value[0]	= '\0';

					if ( RegEnumValue( KeyHandle, Index, Name, &NameLen, NULL, &ValueType, (BYTE*) Value, &ValueSize ) != ERROR_SUCCESS )
						break;

					Match = FALSE;

					if ( !wcsncmp( Name, RegMacroPrefix, wcslen( RegMacroPrefix ) ) ) {
						if ( iswdigit( Name[ wcslen( RegMacroPrefix ) ] ) ) {
							if ( !ParseInt( Name + wcslen( RegMacroPrefix ), &Number, &sep1 ) )
								if ( *sep1 == '\0' )
									Match = TRUE;
						}
						else if ( iswalpha( Name[ wcslen( RegMacroPrefix ) ] ) ) {
							if ( IsValidUserMacroName( Name + wcslen( RegMacroPrefix ) ) )
								Match = TRUE;
						}
					}

					if ( Match && MatchEntries )
						Match = MacroIsInList( Name + wcslen( RegMacroPrefix ), MatchEntries );


					if ( Match ) {
						if ( iswalpha( Name[ wcslen( RegMacroPrefix ) ] ) ) {			//	add string-named macro to list
							MacroEntry[Count].Number = 0;
							wcscpy_s( MacroEntry[Count].MacroName, MAX_MACRO_NAME_LEN + 1, Name + wcslen( RegMacroPrefix ) );
							NameLen = (unsigned int) ( wcslen( Name ) - wcslen( RegMacroPrefix ) );
						}
						else if ( iswdigit( Name[ wcslen( RegMacroPrefix ) ] ) ) {		//	add numeric macro to list
							MacroEntry[Count].Number = Number;
							MacroEntry[Count].MacroName[0] ='\0';
							NameLen = DigitsIn( Number );
						}

						wcscpy_s( MacroEntry[Count].Value, NI_MAXHOST, Value );

						if ( NameLen > LongestName )
							LongestName = NameLen;

						Count++;
					}
				}



				if ( Count > 1 )
					qsort( MacroEntry, Count, sizeof( _SortMacroEntry ), MacroCmp$ );


				for ( int i = 0; i < Count; i++ ) {

					if ( MacroEntry[i].MacroName[0] ) {
						Printf( L"#%s ", MacroEntry[i].MacroName );
						ShowSpaces( LongestName - (unsigned int) wcslen( MacroEntry[i].MacroName ) );
					}
					else {
						Printf( L"#%u ", MacroEntry[i].Number );
						ShowSpaces( LongestName - DigitsIn( MacroEntry[i].Number ) );
					}

					if ( HLprintf( L"= \x12%s\r\n", MacroEntry[i].Value ) == USER_ABORT )
						goto Abort;
				}

				free( MacroEntry );
				MacroEntry = NULL;
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error allocating sort buffer for macros.\r\n" );
				CL_ReleaseBreak();
				return -1;
			}
		}

		RegCloseKey( KeyHandle );
	}


	if ( !Count ) {
		if ( HLprintf( L"No macros defined.\r\n" ) == USER_ABORT )
			goto Abort;
	}
	else if ( ShowCount ) {
		if ( HLprintf( L"\r\n   %d %s%s defined.\r\n", Count, MatchEntries ? L"matching " : L"" ,( Count == 1 ) ? L"macro" : L"macros" ) == USER_ABORT )
			goto Abort;
	}

	CL_ReleaseBreak();
	return Count;


Abort:

	if ( MacroEntry )
		free( MacroEntry );

	CL_ReleaseBreak();
	return USER_ABORT;
}



unsigned int GetDefaultTTL()
{
	MIB_IPSTATS IPStats;


	memset( &IPStats, 0x00, sizeof( IPStats ) );
	GetIpStatistics( &IPStats );

	if ( ( IPStats.dwDefaultTTL >= TTL_MIN ) && ( IPStats.dwDefaultTTL <= TTL_MAX ) )
		return IPStats.dwDefaultTTL;
	else
		return TTL_DEFAULT;
}



DWORD GetDnsAddress( unsigned int Index )
{
	FIXED_INFO *NetworkInfo		= NULL;
	ULONG Size					= 0;
	IP_ADDR_STRING *This		= NULL;

	DWORD DnsAddress			= 0;
	DWORD Address				= 0;
	unsigned int Count			= 0;


	GetNetworkParams( NULL, &Size );

	if ( Size ) {
		NetworkInfo = (FIXED_INFO*) malloc( Size );

		if ( NetworkInfo ) {
			if ( GetNetworkParams( NetworkInfo, &Size ) == ERROR_SUCCESS ) {

				This = &(NetworkInfo->DnsServerList);

				while ( This ) {

					if ( This->IpAddress.String[0] ) {
						DnsAddress = 0;
						DnsAddress = inet_addr( This->IpAddress.String );

						if ( ( DnsAddress != 0 ) && ( DnsAddress != INADDR_NONE ) ) {
							Count++;
							if ( Count == Index )
								Address = DnsAddress;
						}
					}

					This = This->Next;
				}
			}

			free( NetworkInfo );
		}
	}
	
	if ( !Index )
		return Count;
	else
		return Address;
}



int GetDnsAddress( sockaddr_in *Address, unsigned int Index )
{
	if ( !Address )
		return -1;


	DWORD DnsAddress			= 0;
	int rv						= 1;


	DnsAddress = GetDnsAddress( Index );

	if ( DnsAddress ) {
		Address->sin_addr.S_un.S_addr = DnsAddress;
		Address->sin_family = AF_INET;
		Address->sin_port = 0;
		rv = 0;
	}

	return rv;
}



DWORD GetDhcpAddress( unsigned int Index )
{
	IP_ADAPTER_INFO *AdaptersInfo	= NULL;
	ULONG Size						= NULL;
	IP_ADAPTER_INFO *This			= NULL;

	DWORD DhcpAddress				= 0;
	DWORD Address					= 0;
	unsigned int Count				= 0;


	GetAdaptersInfo( NULL, &Size );

	if ( Size ) {
		AdaptersInfo = (IP_ADAPTER_INFO*) malloc( Size );

		if ( AdaptersInfo ) {
			if ( GetAdaptersInfo( AdaptersInfo, &Size ) == ERROR_SUCCESS ) {

				This = AdaptersInfo;

				while ( This ) {

					if ( This->DhcpEnabled ) {
						if ( This->DhcpServer.IpAddress.String[0] ) {
							DhcpAddress = 0;
							DhcpAddress = inet_addr( This->DhcpServer.IpAddress.String );

							if ( ( DhcpAddress != 0 ) && ( DhcpAddress != INADDR_NONE ) ) {
								Count++;
								if ( Count == Index )
									Address = DhcpAddress;
							}
						}
					}

					This = This->Next;
				}
			}

			free( AdaptersInfo );
		}
	}
	
	if ( !Index )
		return Count;
	else
		return Address;
}



int GetDhcpAddress( sockaddr_in *Address, unsigned int Index )
{
	if ( !Address )
		return -1;


	DWORD DhcpAddress			= 0;
	int rv						= 1;


	DhcpAddress = GetDhcpAddress( Index );

	if ( DhcpAddress ) {
		Address->sin_addr.S_un.S_addr = DhcpAddress;
		Address->sin_family = AF_INET;
		Address->sin_port = 0;
		rv = 0;
	}

	return rv;
}



DWORD GetWinsAddress( unsigned int Index )
{
	IP_ADAPTER_INFO *AdaptersInfo		= NULL;
	ULONG Size							= 0;
	IP_ADAPTER_INFO *This				= NULL;

	DWORD WinsAddress					= 0;
	DWORD Address						= 0;
	unsigned int Count					= 0;


	GetAdaptersInfo( NULL, &Size );

	if ( Size ) {
		AdaptersInfo = (IP_ADAPTER_INFO*) malloc( Size );

		if ( AdaptersInfo ) {
			if ( GetAdaptersInfo( AdaptersInfo, &Size ) == ERROR_SUCCESS ) {

				This = AdaptersInfo;

				while ( This ) {

					if ( This->HaveWins ) {

						if ( This->PrimaryWinsServer.IpAddress.String[0] ) {
							WinsAddress = 0;
							WinsAddress = inet_addr( This->PrimaryWinsServer.IpAddress.String );

							if ( ( WinsAddress != 0 ) && ( WinsAddress != INADDR_NONE ) ) {
								Count++;
								if ( Count == Index )
									Address = WinsAddress;
							}
						}

						if ( This->SecondaryWinsServer.IpAddress.String[0] ) {
							WinsAddress = 0;
							WinsAddress = inet_addr( This->SecondaryWinsServer.IpAddress.String );

							if ( ( WinsAddress != 0 ) && ( WinsAddress != INADDR_NONE ) ) {
								Count++;
								if ( Count == Index )
									Address = WinsAddress;
							}
						}
					}

					This = This->Next;
				}
			}
			
			free( AdaptersInfo );
		}
	}
	
	if ( !Index )
		return Count;
	else
		return Address;
}



int GetWinsAddress( sockaddr_in *Address, unsigned int Index )
{
	if ( !Address )
		return -1;


	DWORD WinsAddress			= 0;
	int rv						= 1;


	WinsAddress = GetWinsAddress( Index );

	if ( WinsAddress ) {
		Address->sin_addr.S_un.S_addr = WinsAddress;
		Address->sin_family = AF_INET;
		Address->sin_port = 0;
		rv = 0;
	}

	return rv;
}



DWORD GetRouterAddress( unsigned int Index )
{
	IP_ADAPTER_INFO *AdaptersInfo		= NULL;
	ULONG Size							= 0;
	IP_ADAPTER_INFO *This				= NULL;
	IP_ADDR_STRING *GWEntry				= NULL;

	DWORD RouterAddress					= 0;
	DWORD Address						= 0;
	unsigned int Count					= 0;


	GetAdaptersInfo( NULL, &Size );

	if ( Size ) {
		AdaptersInfo = (IP_ADAPTER_INFO*) malloc( Size );

		if ( AdaptersInfo ) {
			if ( GetAdaptersInfo( AdaptersInfo, &Size ) == ERROR_SUCCESS ) {

				This = AdaptersInfo;

				while ( This ) {

					GWEntry = &( This->GatewayList );

					while ( GWEntry ) {
						if ( GWEntry->IpAddress.String[0] ) {
							RouterAddress = 0;
							RouterAddress = inet_addr( GWEntry->IpAddress.String );

							if ( ( RouterAddress != 0 ) && ( RouterAddress != INADDR_NONE ) ) {
								Count++;
								if ( Count == Index )
									Address = RouterAddress;
							}
						}

						GWEntry = GWEntry->Next;
					}

					This = This->Next;
				}
			}

			free( AdaptersInfo );
		}
	}
	
	if ( !Index )
		return Count;
	else
		return Address;
}



int GetRouterAddress( sockaddr_in *Address, unsigned int Index )
{
	if ( !Address )
		return -1;


	DWORD RouterAddress			= 0;
	int rv						= 1;


	RouterAddress = GetRouterAddress( Index );

	if ( RouterAddress ) {
		Address->sin_addr.S_un.S_addr = RouterAddress;
		Address->sin_family = AF_INET;
		Address->sin_port = 0;
		rv = 0;
	}

	return rv;
}



int GetMacroValue( unsigned int Number, LPTSTR outValue )
{
	if ( !outValue )
		return -1;


	wchar_t RegValueName[SHORT_BUF_LEN]		= L"";
	wchar_t RegData[NI_MAXHOST]				= L"";
	HKEY KeyHandle							= NULL;
	DWORD ValueType							= 0;
	DWORD ValueSize							= NI_MAXHOST * sizeof( wchar_t );

	int rv									= 1;

	
	wsprintf( RegValueName, L"%s%u", RegMacroPrefix, Number );

	if ( RegOpenKeyEx( HKEY_CURRENT_USER, RegKeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
		if ( RegQueryValueEx( KeyHandle, RegValueName, NULL, &ValueType, (BYTE*) RegData, &ValueSize ) == ERROR_SUCCESS ) {
			if ( RegData[0] && ( ValueType == REG_SZ ) ) {
				wcscpy_s( outValue, NI_MAXHOST, RegData );
				rv = 0;
			}
		}

		RegCloseKey( KeyHandle );
	}

	return rv;
}



int GetUserMacroValue( LPTSTR MacroName, LPTSTR outValue )
{
	if ( !MacroName || !outValue )
		return -1;
	else if ( !iswalpha( MacroName[0] ) )
		return -1;


	wchar_t RegValueName[SHORT_BUF_LEN]		= L"";
	wchar_t RegData[NI_MAXHOST]				= L"";
	HKEY KeyHandle							= NULL;
	DWORD ValueType							= 0;
	DWORD ValueSize							= NI_MAXHOST * sizeof( wchar_t );

	int rv			= 1;


	wcscpy_s( RegValueName, SHORT_BUF_LEN, RegMacroPrefix );
	wcscat_s( RegValueName, SHORT_BUF_LEN, MacroName );
	_wcsupr_s( RegValueName + wcslen( RegMacroPrefix ), SHORT_BUF_LEN );

	if ( RegOpenKeyEx( HKEY_CURRENT_USER, RegKeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
		if ( RegQueryValueEx( KeyHandle, RegValueName, NULL, &ValueType, (BYTE*) RegData, &ValueSize ) == ERROR_SUCCESS ) {
			if ( RegData[0] && ( ValueType == REG_SZ ) ) {
				wcscpy_s( outValue, NI_MAXHOST, RegData );
				rv = 0;
			}
		}

		RegCloseKey( KeyHandle );
	}

	return rv;
}



int GetMacroValue( LPTSTR HostName, LPTSTR outValue )
{
	if ( !HostName || !outValue )
		return -1;

	if ( HostName[0] != '#' )
		return 1;


	unsigned int Number			= 0;
	wchar_t *sep1				= NULL;


	if ( iswalpha( HostName[1] ) )
		if ( IsValidUserMacroName( HostName + 1 ) )
			return GetUserMacroValue( HostName + 1, outValue );

	if ( ParseInt( HostName + 1, &Number, &sep1 ) )
		return 1;

	if ( *sep1 != '\0' )
		return 1;

	return GetMacroValue( Number, outValue );
}



int ResolveMagicNames( LPTSTR HostName, sockaddr_in *AddressV4, sockaddr_in6 *AddressV6 )
{
	if ( !HostName || !AddressV4 || !AddressV6 )
		return -1;

	if ( HostName[0] == '\0' )
		return 0;


	unsigned int Index		= 0;
	int rv					= 0;


	if ( Index = IsMagicName( HostName, MagicName[MAGIC_ROUTER_NAME] ) )
		GetRouterAddress( AddressV4, Index );
	else if ( Index = IsMagicName( HostName, MagicName[MAGIC_ROUTER_NAME_2] ) )
		GetRouterAddress( AddressV4, Index );
	else if ( Index = IsMagicName( HostName, MagicName[MAGIC_DNS_NAME] ) )
		GetDnsAddress( AddressV4, Index );
	else if ( Index = IsMagicName( HostName, MagicName[MAGIC_DHCP_NAME] ) )
		GetDhcpAddress( AddressV4, Index );
	else if ( Index = IsMagicName( HostName, MagicName[MAGIC_WINS_NAME] ) )
		GetWinsAddress( AddressV4, Index );

	if ( AddressV4->sin_family == AF_INET )
		rv = 1;

	return rv;
}



BOOL IpV6AddrNotNull( IP6_ADDRESS Address )
{
	for ( int i = 0; i < 4; i++ )
		if ( Address.IP6Dword[i] )
			return TRUE;

	return FALSE;
}



int LookupHostIP( LPTSTR HostName, SOCKADDR_STORAGE *HostIP, BYTE TryFirst, BYTE TrySecond, LPTSTR outCName, BOOL *outIsMacro )
{
	if ( !HostName || !HostIP )
		return -1;


	//		I make two separate calls to GetAddrInfoW(), once for IPv4, then again for IPv6.
	//		When I use just one call using AF_UNSPEC, I sometimes get just IPv4 and any
	//		IPv6 address is lost.  Even if only one protocol is specified, I try both to
	//		maximize my chances of picking up the CNAME field.


	//	char HostNameA[2048]				= "";
	//	hostent *HostEntry					= NULL;
	//	DWORD *DWordPtr						= NULL;

	addrinfoW AddressHints;
	addrinfoW *AddressInfo				= NULL;
	addrinfoW *This						= NULL;
	sockaddr_in AddressV4;
	sockaddr_in6 AddressV6;
	wchar_t CName[NI_MAXHOST]			= L"";
	wchar_t MacroValue[NI_MAXHOST]		= L"";
	DNS_RECORD *DnsResults				= NULL;
	DNS_RECORD *ThisDns					= NULL;

	int rv								= 0;


	if ( !HostName || !HostIP )
		return -1;

	if ( ( TryFirst != AF_INET ) && ( TryFirst != AF_INET6 ) )
		return -2;


	memset( &AddressV4, 0x00, sizeof( AddressV4 ) );
	memset( &AddressV6, 0x00, sizeof( AddressV6 ) );


	ResolveMagicNames( HostName, &AddressV4, &AddressV6 );

	if ( AddressV4.sin_family == 0 )
		if ( !GetMacroValue( HostName, MacroValue ) )
			HostName = MacroValue;


	if ( AddressV4.sin_family == 0 ) {

		memset( &AddressHints, 0x00, sizeof( AddressHints ) );
		AddressHints.ai_family = AF_INET;
		if ( outCName )
			AddressHints.ai_flags |= AI_CANONNAME;

		GetAddrInfoW( HostName, NULL, &AddressHints, &AddressInfo );

		if ( AddressInfo ) {
			This = AddressInfo;

			while ( This ) {
				//	wprintf( L"   * This = 0x%08X   Next = 0x%08X\n", This, This->ai_next );
				//	wprintf( L"   * AddressFamily = %u   AddrLen = %u\n", This->ai_family, This->ai_addrlen );
				//	if ( This->ai_canonname )
				//		wprintf( L"   * CanonName = '%s'\n", This->ai_canonname );

				if ( ( This->ai_family == AF_INET ) && ( This->ai_addrlen == sizeof( sockaddr_in ) ) && This->ai_addr && ( AddressV4.sin_family == 0 ) )
					AddressV4 = *( (sockaddr_in*) This->ai_addr );

				if ( This->ai_canonname && ( CName[0] == '\0' ) )
					wcscpy_s( CName, NI_MAXHOST, This->ai_canonname );

				This = This->ai_next;
			}

			FreeAddrInfoW( AddressInfo );
			AddressInfo = NULL;
		}
	}


	memset( &AddressHints, 0x00, sizeof( AddressHints ) );
	AddressHints.ai_family = AF_INET6;
	if ( outCName && ( CName[0] == '\0' ) )
		AddressHints.ai_flags |= AI_CANONNAME;

	//	wprintf( L"   * Try IPv6:\n" );
	GetAddrInfoW( HostName, NULL, &AddressHints, &AddressInfo );

	if ( AddressInfo ) {
		This = AddressInfo;

		while ( This ) {
			//	wprintf( L"   * This = 0x%08X   Next = 0x%08X\n", This, This->ai_next );
			//	wprintf( L"   * AddressFamily = %u   AddrLen = %u\n", This->ai_family, This->ai_addrlen );
			//	if ( This->ai_canonname )
			//		wprintf( L"   * CanonName = '%s'\n", This->ai_canonname );

			if ( ( This->ai_family == AF_INET6 ) && ( This->ai_addrlen == sizeof( sockaddr_in6 ) ) && This->ai_addr && ( AddressV6.sin6_family == 0 ) )
				AddressV6 = *( (sockaddr_in6*) This->ai_addr );

			if ( This->ai_canonname && ( CName[0] == '\0' ) )
				wcscpy_s( CName, NI_MAXHOST, This->ai_canonname );

			This = This->ai_next;
		}

		FreeAddrInfoW( AddressInfo );
		AddressInfo = NULL;
	}
	

	if ( AddressV6.sin6_family == 0 ) {
		//		GetAddrInfoW() let us down; let's try DnsQuery_W() :

		if ( ( DnsQuery_W( HostName, DNS_TYPE_AAAA, DNS_QUERY_STANDARD, NULL, &DnsResults, NULL ) == 0 ) && DnsResults ) {
			ThisDns = DnsResults;

			while ( ThisDns ) {
				if ( ThisDns->wType == DNS_TYPE_AAAA ) {
					if ( IpV6AddrNotNull( ThisDns->Data.AAAA.Ip6Address ) ) {
						memcpy( &( AddressV6.sin6_addr ), &( ThisDns->Data.AAAA.Ip6Address ), 16 );
						AddressV6.sin6_family = AF_INET6;
						break;
					}
				}

				ThisDns = ThisDns->pNext;
			}

			DnsRecordListFree( DnsResults, DnsFreeRecordsList );
		}
	}


	if ( ( TryFirst == AF_INET ) && ( AddressV4.sin_family == AF_INET ) )
		*( (sockaddr_in*) HostIP ) = AddressV4;
	else if ( ( TryFirst == AF_INET6 ) && ( AddressV6.sin6_family == AF_INET6 ) )
		*( (sockaddr_in6*) HostIP ) = AddressV6;
	else if ( ( TrySecond == AF_INET ) && ( AddressV4.sin_family == AF_INET ) )
		*( (sockaddr_in*) HostIP ) = AddressV4;
	else if ( ( TrySecond == AF_INET6 ) && ( AddressV6.sin6_family == AF_INET6 ) )
		*( (sockaddr_in6*) HostIP ) = AddressV6;
	else
		rv = 2;

	if ( !rv && CName[0] && outCName )
		wcscpy_s( outCName, NI_MAXHOST, CName );

	if ( outIsMacro ) {
		if ( MacroValue[0] )
			*outIsMacro = TRUE;
		else
			*outIsMacro = FALSE;
	}
	
	return rv;
}



int GetHostName( SOCKADDR_STORAGE *IPAddress, LPTSTR HostName )
{
	if ( !IPAddress || !HostName )
		return -1;

	if ( ( IPAddress->ss_family != AF_INET ) && ( IPAddress->ss_family != AF_INET6 ) )
		return 1;


	wchar_t temp[NI_MAXHOST]			= L"";


	if ( !GetNameInfoW( (SOCKADDR*) IPAddress, sizeof( sockaddr_in6 ), temp, NI_MAXHOST, NULL, 0, 0 ) ) {
		//	wprintf( L"   * GetHostName() returns '%s'\n", temp );
		wcscpy_s( HostName, NI_MAXHOST, temp );
		return 0;
	}

	return 1;
}



int MakeIPv4String( DWORD IPAddress, LPTSTR String )
{
	if ( !String )
		return -1;

	wsprintf( String, L"%u.%u.%u.%u", IPAddress & 0xff, ( IPAddress >> 8 ) & 0xff, ( IPAddress >> 16 ) & 0xff, ( IPAddress >> 24 ) );
	return 0;
}



int MakeIPv6String( sockaddr_in6 *IPAddress, LPTSTR String )
{
	if ( !IPAddress || !String )
		return -1;

	if ( IPAddress->sin6_family != AF_INET6 )
		return 1;


	DWORD BufLen			= IP_STRING_LEN;


	return WSAAddressToString( (LPSOCKADDR) IPAddress, sizeof( sockaddr_in6 ), NULL, String, &BufLen );
}



int MakeIPString( sockaddr *HostAddress, LPTSTR outString )
{
	if ( !HostAddress || !outString )
		return -1;

	if ( ( HostAddress->sa_family != AF_INET ) && ( HostAddress->sa_family != AF_INET6 ) )
		return 1;


	DWORD BufLen			= IP_STRING_LEN;


	return WSAAddressToString( HostAddress, ( HostAddress->sa_family == AF_INET6 ) ? sizeof( sockaddr_in6 ) : sizeof( sockaddr_in ), NULL, outString, &BufLen );
}



BOOL LooksLikeIPv4( LPTSTR String )
{
	if ( !String )
		return FALSE;
	else if ( String[0] == '\0' )
		return FALSE;


	char temp[SHORT_BUF_LEN]	= "";
	DWORD ip					= 0;


	WideCharToMultiByte( CP_ACP, 0, String, -1, temp, SHORT_BUF_LEN, NULL, NULL );
	ip = inet_addr( temp );

	if ( ip && ( ip != INADDR_NONE ) )
		return TRUE;
	else
		return FALSE;
}



int PingTime( HANDLE IcmpHandle, DWORD IPAddress, BYTE *RequestData, unsigned int PacketSize, ICMP_ECHO_REPLY *ReplyBuffer, unsigned int ReplyBufferSize, unsigned int Timeout )
{
	IP_OPTION_INFORMATION *OptionStruct		= (IP_OPTION_INFORMATION*) ( RequestData + PacketSize );
	DWORD PacketsReturned					= 0;
	DWORD LastError							= 0;
	wchar_t ShortBuf[SHORT_BUF_LEN]			= L"";
	int rv									= 0;



	for ( unsigned int i = 0; i < PacketSize; i++ ) {
		unsigned int Value;
		rand_s( &Value );
		RequestData[i] = Value & 0xff;
	}

	memset( ReplyBuffer, 0x00, ReplyBufferSize );


	PacketsReturned = IcmpSendEcho( IcmpHandle, IPAddress, RequestData, PacketSize, OptionStruct, ReplyBuffer, ReplyBufferSize, Timeout );
	LastError = GetLastError();


	if ( PacketsReturned ) {

		PacketsReceived++;
		MakeIPv4String( ReplyBuffer->Address, ShortBuf );

		HLprintf( L"Reply from \x12%s\x11: ", ShortBuf );

		if ( ReplyBuffer->Status == IP_SUCCESS ) {

			Printf( L"bytes=%u ", ReplyBuffer->DataSize );

			if ( ReplyBuffer->RoundTripTime )
				Printf( L"time=%ums TTL=%u\r\n", ReplyBuffer->RoundTripTime, ReplyBuffer->Options.Ttl );
			else
				Printf( L"time<1ms TTL=%u\r\n", ReplyBuffer->Options.Ttl );

			rv = ReplyBuffer->RoundTripTime;

			if ( memcmp( RequestData, ReplyBuffer->Data, PacketSize ) ) {
				QPuts( L"    Data mismatch!\r\n" );
				rv = ERR_DATA_MISMATCH;
			}
		}
		else if ( ReplyBuffer->Status == IP_DEST_NET_UNREACHABLE ) {
			QPuts( L"Destination net unreachable.\r\n" );
			rv = - ( IP_DEST_NET_UNREACHABLE );
		}
		else if ( ReplyBuffer->Status == IP_DEST_HOST_UNREACHABLE ) {
			QPuts( L"Destination host unreachable.\r\n" );
			rv = - ( IP_DEST_HOST_UNREACHABLE );
		}
		else if ( ReplyBuffer->Status == IP_DEST_PROT_UNREACHABLE ) {
			QPuts( L"Destination protocol unreachable.\r\n" );
			rv = - ( IP_DEST_PROT_UNREACHABLE );
		}
		else if ( ReplyBuffer->Status == IP_DEST_PORT_UNREACHABLE ) {
			QPuts( L"Destination port unreachable.\r\n" );
			rv = - ( IP_DEST_PORT_UNREACHABLE );
		}
		else if ( ReplyBuffer->Status == IP_TTL_EXPIRED_TRANSIT ) {
			QPuts( L"TTL expired in transit.\r\n" );
			rv = - ( IP_TTL_EXPIRED_TRANSIT );
		}
		else {
			Printf( L"Status = %u\r\n", ReplyBuffer->Status );
			rv = ERR_UNKNOWN;
		}

		TotalPingTime += ReplyBuffer->RoundTripTime;

		if ( ReplyBuffer->RoundTripTime < MinPingTime )
			MinPingTime = ReplyBuffer->RoundTripTime;

		if ( ReplyBuffer->RoundTripTime > MaxPingTime )
			MaxPingTime = ReplyBuffer->RoundTripTime;

	}
	else {
		PacketsLost++;
		ShortBuf[0] = '\0';
		DWORD Size = sizeof( ShortBuf );

		GetIpErrorString( LastError, ShortBuf, &Size );

		if ( ShortBuf[0] )
			Printf( L"%s\r\n", ShortBuf );
		else
			Printf( L"Failed: Error code = %u\r\n", LastError );

		rv = ERR_UNKNOWN;
	}

	return rv;
}



int PingTimeV6( HANDLE IcmpHandle, sockaddr_in6 *IPAddress, BYTE *RequestData, unsigned int PacketSize, ICMPV6_ECHO_REPLY *ReplyBuffer, unsigned int ReplyBufferSize, unsigned int Timeout )
{
	IP_OPTION_INFORMATION *OptionStruct		= (IP_OPTION_INFORMATION*) ( RequestData + PacketSize );
	sockaddr_in6 SourceAddress;
	sockaddr_in6 ReplyFrom;
	DWORD PacketsReturned					= 0;
	DWORD LastError							= 0;
	wchar_t ShortBuf[SHORT_BUF_LEN]			= L"";
	BYTE *ReplyData							= (BYTE*) ReplyBuffer + sizeof( ICMPV6_ECHO_REPLY );
	int rv									= 0;



	for ( unsigned int i = 0; i < PacketSize; i++ ) {
		unsigned int Value;
		rand_s( &Value );
		RequestData[i] = Value & 0xff;
	}

	memset( ReplyBuffer, 0x00, ReplyBufferSize );

	memset( &SourceAddress, 0x00, sizeof( sockaddr_in6 ) );
	SourceAddress.sin6_family = AF_INET6;
	SourceAddress.sin6_addr = in6addr_any;


	
	PacketsReturned = Icmp6SendEcho2( IcmpHandle, NULL, NULL, NULL, &SourceAddress, IPAddress, RequestData, PacketSize, OptionStruct, ReplyBuffer, ReplyBufferSize, Timeout );
	LastError = GetLastError();


	if ( PacketsReturned ) {

		PacketsReceived++;

		memset( &ReplyFrom, 0x00, sizeof( sockaddr_in6 ) );
		ReplyFrom.sin6_family = AF_INET6;
		for ( unsigned int j = 0; j < 8; j++ )
			ReplyFrom.sin6_addr.u.Word[j] = ReplyBuffer->Address.sin6_addr[j];
		ReplyFrom.sin6_scope_id = ReplyBuffer->Address.sin6_scope_id;
		//	ReplyFrom.sin6_scope_id = IPAddress->sin6_scope_id;

		MakeIPv6String( &ReplyFrom, ShortBuf );

		HLprintf( L"Reply from \x12%s\x11: ", ShortBuf );

		if ( ReplyBuffer->Status == IP_SUCCESS ) {

			if ( ReplyBuffer->RoundTripTime )
				Printf( L"time=%ums\r\n", ReplyBuffer->RoundTripTime );
			else
				Printf( L"time<1ms\r\n" );

			rv = ReplyBuffer->RoundTripTime;

			if ( memcmp( RequestData, ReplyData, PacketSize ) ) {
				QPuts( L"    Data mismatch!\n" );
				rv = ERR_DATA_MISMATCH;
			}
		}
		else if ( ReplyBuffer->Status == IP_DEST_NET_UNREACHABLE ) {
			QPuts( L"Destination net unreachable.\r\n" );
			rv = - ( IP_DEST_NET_UNREACHABLE );
		}
		else if ( ReplyBuffer->Status == IP_DEST_HOST_UNREACHABLE ) {
			QPuts( L"Destination host unreachable.\r\n" );
			rv = - ( IP_DEST_HOST_UNREACHABLE );
		}
		else if ( ReplyBuffer->Status == IP_DEST_PROT_UNREACHABLE ) {
			QPuts( L"Destination protocol unreachable.\r\n" );
			rv = - ( IP_DEST_PROT_UNREACHABLE );
		}
		else if ( ReplyBuffer->Status == IP_DEST_PORT_UNREACHABLE ) {
			QPuts( L"Destination port unreachable.\r\n" );
			rv = - ( IP_DEST_PORT_UNREACHABLE );
		}
		else if ( ReplyBuffer->Status == IP_HOP_LIMIT_EXCEEDED ) {
			QPuts( L"Hop limit exceeded in transit.\r\n" );
			rv = - ( IP_HOP_LIMIT_EXCEEDED );
		}
		else {
			Printf( L"Status = %u\r\n", ReplyBuffer->Status );
			rv = ERR_UNKNOWN;
		}

		TotalPingTime += ReplyBuffer->RoundTripTime;

		if ( ReplyBuffer->RoundTripTime < MinPingTime )
			MinPingTime = ReplyBuffer->RoundTripTime;

		if ( ReplyBuffer->RoundTripTime > MaxPingTime )
			MaxPingTime = ReplyBuffer->RoundTripTime;

	}
	else {
		PacketsLost++;
		ShortBuf[0] = '\0';
		DWORD Size = sizeof( ShortBuf );

		GetIpErrorString( LastError, ShortBuf, &Size );

		if ( ShortBuf[0] )
			Printf( L"%s\r\n", ShortBuf );
		else
			Printf( L"Failed: Error code = %u\r\n", LastError );

		rv = ERR_UNKNOWN;
	}

	return rv;
}



int ShowPingStatistics( LPTSTR HostIPString )
{
	if ( ( PacketsReceived | PacketsLost ) == 0 )
		return 0;


	unsigned int PacketsSent = PacketsReceived + PacketsLost;
	unsigned int LossPercent = PacketsLost * 100 / PacketsSent;


	Printf( L"\r\nPing statistics for %s:\r\n", HostIPString );
	Printf( L"    Packets: Sent = %u, Received = %u, Lost = %u (%u%% loss)\r\n", PacketsSent, PacketsReceived, PacketsLost, LossPercent );

	if ( PacketsReceived ) {
		Printf( L"Approximate round trip times in milliseconds:\r\n" );
		Printf( L"    Minimum = %ums, Maximum = %ums, Average = %ums\r\n", MinPingTime, MaxPingTime, TotalPingTime / PacketsReceived );
	}

	return 0;
}



BOOL WINAPI ConsoleCtrlHandler( DWORD EventType )
{
	if ( ( EventType == CTRL_C_EVENT ) || ( EventType == CTRL_BREAK_EVENT ) ) {

		if ( EventType == CTRL_C_EVENT )
			ConsoleCtrlFlags |= FLAG_CONTROL_C;
		else if ( EventType == CTRL_BREAK_EVENT )
			ConsoleCtrlFlags |= FLAG_CONTROL_BREAK;

		return TRUE;
	}

	return FALSE;
}



int ConsoleHandle( HANDLE inHandle ) {
	if ( !inHandle )
		return 0;

	if ( ( inHandle == (HANDLE) STD_INPUT_HANDLE ) || ( inHandle == (HANDLE) STD_OUTPUT_HANDLE ) || ( inHandle == (HANDLE) STD_ERROR_HANDLE ) )
		inHandle = GetStdHandle( (DWORD) inHandle );

	DWORD Mode = UINT_MAX;
	int rv = GetConsoleMode( inHandle, &Mode );

	if ( !rv || ( Mode == UINT_MAX ) )
		return 0;
	else
		return 1;
}



int NewSleep( unsigned int Milliseconds )
{
	unsigned int Interval	= 20;
	DWORD Start				= GetTickCount();
	DWORD Now				= 0;
	DWORD Finish			= Start + Milliseconds;
	BOOL Overflow			= Finish < Start;


	while ( !ConsoleCtrlFlags ) {
		Sleep( Interval );
		Now = GetTickCount();

		if ( Overflow && ( Now > Start ) )
			continue;

		if ( Now >= Finish )
			break;
	}

	return 0;
}



MMRESULT midiOutShortMessage( HMIDIOUT Handle, BYTE Status, BYTE Byte1 = 0, BYTE Byte2 = 0 )
{
	if ( MidiHandle )
		return midiOutShortMsg( Handle, Status | ( Byte1 << 8 ) | ( Byte2 << 16 ) );
	else
		return MMSYSERR_NOERROR;
}



int OpErrorSound( BYTE ErrorDrum )
{
	if ( !Silence && ( ErrorDrum >= MIN_DRUM_KEY ) && ( ErrorDrum <= MAX_DRUM_KEY ) ) {
		midiOutShortMessage( MidiHandle, 0x89, ErrorDrum, 0x00 );
		midiOutShortMessage( MidiHandle, 0x99, ErrorDrum, 0x60 );
		Sleep( 1000 );
	}

	return 0;
}



BOOL IsOption( LPTSTR String, LPTSTR Option, BOOL Suboptions )
{
	if ( !String || !Option )
		return FALSE;

	if ( ( String[0] != '/' ) && ( String[0] != '-' ) )
		return FALSE;

	if ( ( Option[0] == '/' ) || ( Option[0] == '-' ) )
		Option++;

	if ( Suboptions )
		if ( _wcsnicmp( String + 1, Option, wcslen( Option ) ) )
			return FALSE;

	if ( !Suboptions )
		if ( _wcsicmp( String + 1, Option ) )
			return FALSE;

	return TRUE;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI aping( LPTSTR inString )
{
	if ( !inString )
		return -1;


	unsigned int Count				= RegGetDWord( RegKeyName, CountRegName, 0, UINT_MAX, 4 );
	unsigned int Timeout			= RegGetDWord( RegKeyName, TimeoutRegName, 0, PING_MAX_TIMEOUT, PING_DEFAULT_TIMEOUT );
	unsigned int PacketSize			= RegGetDWord( RegKeyName, PacketSizeRegName, PING_MIN_PKTSIZE, PING_MAX_PKTSIZE, PING_DEFAULT_PKTSIZE );
	unsigned int Ttl				= TTL_DEFAULT;
	unsigned int Tos				= 0;
	unsigned int SnoozeTime			= RegGetDWord( RegKeyName, SnoozeTimeRegName, MIN_SNOOZE_TIME, MAX_SNOOZE_TIME, DEFAULT_SNOOZE_TIME );
	unsigned int ErrorSnooze		= RegGetDWord( RegKeyName, ErrorSnoozeRegName, MIN_SNOOZE_TIME, MAX_SNOOZE_TIME, DEFAULT_ERROR_SNOOZE );
	unsigned int SoundVolume		= -1;
	unsigned int SoundMute			= 0;
	unsigned int LastSleep			= 0;

	SOCKADDR_STORAGE HostIP;
	wchar_t HostIPString[IP_STRING_LEN]		= L"";
	wchar_t FullHostName[NI_MAXHOST]		= L"";
	BOOL IsMacro							= FALSE;
	unsigned int NumHostNames				= 0;

	wchar_t DefaultHost[NI_MAXHOST]			= L"";

	unsigned char ArgType[MAX_ARGS];



	HANDLE IcmpHandle4						= INVALID_HANDLE_VALUE;
	HANDLE IcmpHandle6						= INVALID_HANDLE_VALUE;

	BYTE *RequestData						= NULL;
	IP_OPTION_INFORMATION *OptionStruct		= NULL;
	unsigned int ReplyBuffer4Size			= 0;
	unsigned int ReplyBuffer6Size			= 0;
	ICMP_ECHO_REPLY *ReplyBuffer4			= NULL;
	ICMPV6_ECHO_REPLY *ReplyBuffer6			= NULL;
	int PingTimeRv							= 0;

	BYTE TryFirst							= 0;
	BYTE TrySecond							= 0;
	BOOL ResolveAddresses					= FALSE;
	BOOL DontFragment						= FALSE;
	BOOL ShowLineCount						= FALSE;
	BOOL OutputToConsole					= TRUE;
	BOOL NoHighlight						= FALSE;
	int rv									= 1;

	unsigned int Instrument					= RegGetDWord( RegKeyName, InstrumentRegName, MIN_INSTRUMENT, MAX_INSTRUMENT, DEFAULT_INSTRUMENT );
	unsigned int ErrorDrum					= RegGetDWord( RegKeyName, ErrorDrumRegName, 0, MAX_DRUM_KEY, DEFAULT_ERROR_DRUM );
	unsigned int OpErrorDrum				= RegGetDWord( RegKeyName, OpErrorDrumRegName, 0, MAX_DRUM_KEY, OP_ERROR_DRUM );
	unsigned int Octave						= RegGetDWord( RegKeyName, OctaveRegName, 0, 9, 5 );
	unsigned int BaseNote					= 12 * Octave;
	unsigned int Note						= 0;


	if ( ErrorDrum < MIN_DRUM_KEY )
		ErrorDrum = 0;

	if ( OpErrorDrum < MIN_DRUM_KEY )
		OpErrorDrum = 0;


	ConsoleCtrlFlags		= 0;
	Silence					= FALSE;
	Ttl						= GetDefaultTTL();



	if ( !MidiHandle )
		Silence = TRUE;
	else if ( CheckMidiDisabled() )
		Silence = TRUE;



	for ( int i = 0; i < MAX_ARGS; i++ )
		ArgType[i] = 0;


	CL_NoPausing();

	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( IsOption( arg[i], L"/?", FALSE ) )
				return ShowCmdHelp( APingHelpText, FALSE );
			else if ( IsOption( arg[i], L"/??", FALSE ) )
				return ShowCmdHelp( APingHelpText, TRUE );
			else if ( IsOption( arg[i], L"/4", FALSE ) ) {
				if ( !TryFirst )
					TryFirst = AF_INET;
				else if ( TryFirst == AF_INET6 )
					TrySecond = AF_INET;
			}
			else if ( IsOption( arg[i], L"/6", FALSE ) ) {
				if ( !TryFirst )
					TryFirst = AF_INET6;
				else if ( TryFirst == AF_INET )
					TrySecond = AF_INET6;
			}
			else if ( IsOption( arg[i], L"/A", FALSE ) )
				ResolveAddresses = TRUE;
			else if ( IsOption( arg[i], L"/C", FALSE ) )
				ShowLineCount = TRUE;
			else if ( IsOption( arg[i], L"/D", FALSE ) )
				NoHighlight = TRUE;
			else if ( IsOption( arg[i], L"/E", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /E\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &ErrorDrum, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ErrorDrum && ( ( ErrorDrum < MIN_DRUM_KEY ) || ( ErrorDrum > MAX_DRUM_KEY ) ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"ErrorDrum must be 0 or %u to %u: %u\r\n", MIN_DRUM_KEY, MAX_DRUM_KEY, ErrorDrum );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/F", FALSE ) )
				DontFragment = TRUE;
			else if ( IsOption( arg[i], L"/G", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /G\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &Instrument, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( Instrument < MIN_INSTRUMENT ) || ( Instrument > MAX_INSTRUMENT ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Instrument must be %u to %u: %u\r\n", MIN_INSTRUMENT, MAX_INSTRUMENT, Instrument );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/I", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /I\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &Ttl, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( Ttl < TTL_MIN ) || ( Ttl > TTL_MAX ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"TTL must be %u to %u: %u\r\n", TTL_MIN, TTL_MAX, Ttl );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/L", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /L\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &PacketSize, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( PacketSize < PING_MIN_PKTSIZE ) || ( PacketSize > PING_MAX_PKTSIZE ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Packet size must be %u to %u: %u\r\n", PING_MIN_PKTSIZE, PING_MAX_PKTSIZE, PacketSize );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/N", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /N\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &Count, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/O", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /O\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &Octave, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( Octave > 9 ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Octave must be 0 to 9: %u\r\n", Octave );
					return 1;
				}

				BaseNote = 12 * Octave;
			}
			else if ( IsOption( arg[i], L"/Q", FALSE ) )
				Silence = TRUE;
			else if ( IsOption( arg[i], L"/T", FALSE ) )
				Count = 0;
			else if ( IsOption( arg[i], L"/V", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /V\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &Tos, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( Tos > 255 ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"TOS must be 0 to 255: %u\r\n", Tos );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/W", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /W\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &Timeout, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( Timeout < PING_MIN_TIMEOUT ) || ( Timeout > PING_MAX_TIMEOUT ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Timeout must be %u to %u: %u\r\n", PING_MIN_TIMEOUT, PING_MAX_TIMEOUT, Timeout );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/X", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /X\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &OpErrorDrum, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( OpErrorDrum && ( ( OpErrorDrum < MIN_DRUM_KEY ) || ( OpErrorDrum > MAX_DRUM_KEY ) ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"OpErrorDrum must be 0 or %u to %u: %u\r\n", MIN_DRUM_KEY, MAX_DRUM_KEY, OpErrorDrum );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/Y", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /Y\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &SoundVolume, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( SoundVolume > 100 ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Sound volume must be 0 to 100: %u\r\n", SoundVolume );
					return 1;
				}
			}
			else if ( IsOption( arg[i], L"/Z", TRUE ) ) {
				int j = 2;

				if ( ( arg[i][j] == '\0' ) && ( i < ( numargs - 1 ) ) )
					i++, j = 0;
				else if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing value for /Z\r\n" );
					return 1;
				}

				if ( ParseInt( arg[i] + j, &SnoozeTime, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( SnoozeTime < MIN_SNOOZE_TIME ) || ( SnoozeTime > MAX_SNOOZE_TIME ) ) {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"SnoozeTime must be %u to %u: %u\r\n", MIN_SNOOZE_TIME, MAX_SNOOZE_TIME, SnoozeTime );
					return 1;
				}
			}
			else if ( ( arg[i][0] == '/' ) || ( arg[i][0] == '-' ) ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else {
				NumHostNames++;
				ArgType[i] = 1;
			}

		}
	}


	CL_PickHighlightColor ( NoHighlight );


	if ( !NumHostNames ) {
		RegGetString( RegKeyName, DefaultHostRegName, DefaultHost );
		if ( DefaultHost[0] ) {
			NumHostNames = 1;
			arg[0] = DefaultHost;
			arglen[0] = (int) wcslen( DefaultHost );
			ArgType[0] = 1;
			if ( !numargs )
				numargs = 1;
		}
	}

	if ( !NumHostNames ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing host name.\r\n" );
		return 1;
	}


	if ( !TryFirst ) {
		TryFirst = AF_INET;
		TrySecond = AF_INET6;
	}


	if ( ( TryFirst == AF_INET ) || ( TrySecond == AF_INET ) ) {
		IcmpHandle4 = IcmpCreateFile();

		if ( IcmpHandle4 == INVALID_HANDLE_VALUE ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error opening IPv4 echo-request handle.\r\n" );
			OpErrorSound( OpErrorDrum );
			rv = 2;
			goto abort;
		}

		ReplyBuffer4Size = sizeof( ICMP_ECHO_REPLY ) + PacketSize + 16;
		ReplyBuffer4 = (ICMP_ECHO_REPLY*) malloc( ReplyBuffer4Size );

		if ( !ReplyBuffer4 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error allocating IPv4 reply buffer.\r\n" );
			OpErrorSound( OpErrorDrum );
			rv = 2;
			goto abort;
		}
	}

	if ( ( TryFirst == AF_INET6 ) || ( TrySecond == AF_INET6 ) ) {
		IcmpHandle6 = Icmp6CreateFile();

		if ( IcmpHandle6 == INVALID_HANDLE_VALUE ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error opening IPv6 echo-request handle.\r\n" );
			OpErrorSound( OpErrorDrum );
			rv = 2;
			goto abort;
		}

		ReplyBuffer6Size = sizeof( ICMPV6_ECHO_REPLY ) + PacketSize + 16 + sizeof( __IO_STATUS_BLOCK );
		ReplyBuffer6 = (ICMPV6_ECHO_REPLY*) malloc( ReplyBuffer6Size );

		if ( !ReplyBuffer6 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error allocating IPv6 reply buffer.\r\n" );
			OpErrorSound( OpErrorDrum );
			rv = 2;
			goto abort;
		}
	}

	RequestData = (BYTE*) malloc( PacketSize + sizeof( IP_OPTION_INFORMATION ) );

	if ( !RequestData ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Error allocating request data buffer.\r\n" );
		OpErrorSound( OpErrorDrum );
		rv = 2;
		goto abort;
	}

	OptionStruct = (IP_OPTION_INFORMATION*) ( RequestData + PacketSize );
	memset( OptionStruct, 0x00, sizeof( IP_OPTION_INFORMATION ) );
	OptionStruct->Ttl = Ttl;
	OptionStruct->Tos = Tos;
	if ( DontFragment )
		OptionStruct->Flags |= IP_FLAG_DF;



	OutputToConsole = ConsoleHandle( GetStdHandle( STD_OUTPUT_HANDLE ) );

	CL_HookBreak();

	if ( OutputToConsole )
		SetConsoleCtrlHandler( ConsoleCtrlHandler, TRUE );


	if ( ( SoundVolume != -1 ) && !Silence )
		GetSetSoundVolume( ACTION_SET_BOTH, &SoundVolume, &SoundMute );


	if ( !Silence )
		midiOutShortMessage( MidiHandle, 0xc5, Instrument - 1 );			//	select instrument



	//			Loop through specified hostnames:


	for ( int i = 0; i < numargs; i++ ) {
		if ( ArgType[i] == 1 ) {

			PacketsReceived			= 0;
			PacketsLost				= 0;
			MinPingTime				= UINT_MAX;
			MaxPingTime				= 0;
			TotalPingTime			= 0;

			memset( &HostIP, 0x00, sizeof( SOCKADDR_STORAGE ) );
			FullHostName[0] = '\0';
			HostIPString[0] = '\0';
			IsMacro			= FALSE;

			LookupHostIP( arg[i], &HostIP, TryFirst, TrySecond, FullHostName, &IsMacro );

			if ( HostIP.ss_family == 0 ) {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can\u2019t find an address for host %s.\r\n", arg[i] );
				OpErrorSound( OpErrorDrum );
				rv = 2;
			}
			else {
				MakeIPString( (sockaddr*) &HostIP, HostIPString );

				if ( !_wcsicmp( arg[i], FullHostName ) )
					FullHostName[0] = '\0';

				if ( ResolveAddresses && ( FullHostName[0] == '\0' ) )
					GetHostName( &HostIP, FullHostName );
				else if ( ResolveAddresses && IsMacro && LooksLikeIPv4( FullHostName ) )
					GetHostName( &HostIP, FullHostName );

				if ( FullHostName[0] == '\0' )
					wcscpy_s( FullHostName, NI_MAXHOST, arg[i] );

				if ( _wcsicmp( arg[i], HostIPString ) )
					HLprintf( L"\r\nPinging \x12%s\x11 [%s] with %u bytes of data:\r\n\n", FullHostName, HostIPString, PacketSize );
				else
					HLprintf( L"\r\nPinging \x12%s\x11 with %u bytes of data:\r\n\n", FullHostName, PacketSize );


				for ( unsigned int i = 0; ( i < Count ) || !Count; i++ ) {

					ConsoleCtrlFlags = 0;

					if ( ShowLineCount )
						Printf( L"%5u:  ", i + 1 );


					if ( HostIP.ss_family == AF_INET )
						PingTimeRv = PingTime( IcmpHandle4, ( (sockaddr_in*) &HostIP )->sin_addr.S_un.S_addr, RequestData, PacketSize, ReplyBuffer4, ReplyBuffer4Size, Timeout );
					else
						PingTimeRv = PingTimeV6( IcmpHandle6, (sockaddr_in6*) &HostIP, RequestData, PacketSize, (ICMPV6_ECHO_REPLY*) ReplyBuffer6, ReplyBuffer6Size, Timeout );


					if ( PingTimeRv < 0 ) {
						if ( ErrorDrum && !Silence ) {
							midiOutShortMessage( MidiHandle, 0x89, ErrorDrum, 0x00 );
							midiOutShortMessage( MidiHandle, 0x99, ErrorDrum, 0x60 );
						}
						LastSleep = ErrorSnooze;
						NewSleep( ErrorSnooze );
					}
					else {
						if ( rv == 1 )
							rv = 0;

						if ( !Silence ) {
							if ( PingTimeRv == 0 )
								Note = BaseNote + 12;
							else if ( PingTimeRv == 1 )
								Note = BaseNote + 11;
							else if ( PingTimeRv == 2 )
								Note = BaseNote + 9;
							else if ( PingTimeRv <= 5 )
								Note = BaseNote + 7;
							else if ( PingTimeRv <= 10 )
								Note = BaseNote + 5;
							else if ( PingTimeRv <= 50 )
								Note = BaseNote + 4;
							else if ( PingTimeRv <= 100 )
								Note = BaseNote + 2;
							else
								Note = BaseNote;
					
							// wprintf( L"   * Play note %u\n", Note );
							midiOutShortMessage( MidiHandle, 0x85, Note, 0x00 );
							midiOutShortMessage( MidiHandle, 0x95, Note, 0x60 );		//	play a note
						}

						LastSleep = SnoozeTime;
						NewSleep( SnoozeTime );
					}

					if ( ConsoleCtrlFlags ) {
						ShowPingStatistics( HostIPString );

						if ( ConsoleCtrlFlags & FLAG_CONTROL_BREAK )
							QPuts( L"\r\n" );
					}

					if ( ConsoleCtrlFlags & FLAG_CONTROL_C )
						break;

				}					//	end number-of-pings loop


				if ( ( ConsoleCtrlFlags & FLAG_CONTROL_C ) == 0 )
					ShowPingStatistics( HostIPString );

			}

		}
	}							//	end iterate-through-arguments loop


abort:

	free( RequestData );
	free( ReplyBuffer4 );
	free( ReplyBuffer6 );
	IcmpCloseHandle( IcmpHandle4 );
	IcmpCloseHandle( IcmpHandle6 );

	CL_ReleaseBreak();

	if ( OutputToConsole )
		SetConsoleCtrlHandler( ConsoleCtrlHandler, FALSE );

	if ( !Silence ) {
		if ( LastSleep < 1750 )
			Sleep( 1750 - LastSleep );

		//	midiOutReset( MidiHandle );
		midiOutShortMessage( MidiHandle, 0xb5, 123, 0 );		//	all notes off
		midiOutShortMessage( MidiHandle, 0xb9, 123, 0 );
	}

	return rv;
}



DLLExports int WINAPI _pingavg( LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( PacketsReceived )
		wsprintf( outString, L"%u", TotalPingTime / PacketsReceived );
	else
		wcscpy_s( outString, MAX_ARG_LEN, L"00" );

	return 0;
}



DLLExports int WINAPI _pingmax( LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( PacketsReceived )
		wsprintf( outString, L"%u", MaxPingTime );
	else
		wcscpy_s( outString, MAX_ARG_LEN, L"00" );

	return 0;
}



DLLExports int WINAPI _pingmin( LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( PacketsReceived )
		wsprintf( outString, L"%u", MinPingTime );
	else
		wcscpy_s( outString, MAX_ARG_LEN, L"00" );

	return 0;
}



DLLExports int WINAPI _pingslost( LPTSTR outString )
{
	if ( !outString )
		return -1;


	wsprintf( outString, L"%u", PacketsLost );
	return 0;
}



DLLExports int WINAPI _pingsrcvd( LPTSTR outString )
{
	if ( !outString )
		return -1;


	wsprintf( outString, L"%u", PacketsReceived );
	return 0;
}



DLLExports int WINAPI f_pingdns( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int Index			= 1;
	DWORD Address				= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &Index, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index value: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}


	Address = GetDnsAddress( Index );

	if ( !Index )
		wsprintf( String, L"%u", Address );
	else if ( !Address )
		String[0] = '\0';
	else
		MakeIPv4String( Address, String );
	
	return 0;
}



DLLExports int WINAPI f_pingdhcp( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int Index			= 1;
	DWORD Address				= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &Index, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index value: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}


	Address = GetDhcpAddress( Index );

	if ( !Index )
		wsprintf( String, L"%u", Address );
	else if ( !Address )
		String[0] = '\0';
	else
		MakeIPv4String( Address, String );

	return 0;
}



DLLExports int WINAPI f_pingrouter( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int Index			= 1;
	DWORD Address				= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &Index, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index value: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}


	Address = GetRouterAddress( Index );

	if ( !Index )
		wsprintf( String, L"%u", Address );
	else if ( !Address )
		String[0] = '\0';
	else
		MakeIPv4String( Address, String );

	return 0;
}



DLLExports int WINAPI f_pingwins( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int Index			= 1;
	DWORD Address				= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &Index, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index value: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}


	Address = GetWinsAddress( Index );

	if ( !Index )
		wsprintf( String, L"%u", Address );
	else if ( !Address )
		String[0] = '\0';
	else
		MakeIPv4String( Address, String );

	return 0;
}



DLLExports int WINAPI f_lookuphost( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int Flags		= 0;
	wchar_t *HostName		= String;
	wchar_t *sep1			= String;
	BYTE TryFirst			= AF_INET;
	BYTE TrySecond			= AF_INET6;

	SOCKADDR_STORAGE HostIP;
	wchar_t CName[NI_MAXHOST]	= L"";


	while ( iswspace( *HostName ) )
		HostName++;

	if ( iswdigit( *HostName ) ) {
		
		if ( !ParseInt( HostName, &Flags, &sep1 ) ) {
			if ( ( *sep1 == ',' ) || ( *sep1 == ';' ) ) {
				HostName = sep1 + 1;

				while ( iswspace( *HostName ) )
					HostName++;
			}
			else
				Flags = 0;
		}
	}

	if ( *HostName == '\0' ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Missing hostname\r\n" );
		return -1;
	}


	memset( &HostIP, 0x00, sizeof( SOCKADDR_STORAGE ) );

	switch ( Flags & 127 ) {

	case 4 :
		TryFirst = AF_INET;
		TrySecond = 0;
		break;

	case 6 :
		TryFirst = AF_INET6;
		TrySecond = 0;
		break;

	case 10 :
		TryFirst = AF_INET6;
		TrySecond = AF_INET;
		break;

	default :
		TryFirst = AF_INET;
		TrySecond = AF_INET6;
	}


	if ( LookupHostIP( HostName, &HostIP, TryFirst, TrySecond, CName, NULL ) ) {
		String[0] = '\0';
		return 0;
	}

	String[0] = '\0';

	if ( Flags & 128 ) {
		wcscpy_s( String, MAX_ARG_LEN, CName );
		return 0;
	}

	MakeIPString( (sockaddr*) &HostIP, String );
	return 0;
}



DLLExports int WINAPI f_apingmacro( LPTSTR String )
{
	if ( !String )
		return -1;


	wchar_t temp[MAX_ARG_LEN]	= L"";

	unsigned int Number			= 0;
	wchar_t *ch					= String;
	int rv						= 0;


	while ( iswspace( *ch ) )
		ch++;

	if ( *ch == '#' )
		ch++;

	if ( iswalpha( *ch ) ) {
		if ( IsValidUserMacroName( ch ) ) {
			wcscpy_s( temp, MAX_ARG_LEN, ch );
			_wcsupr_s( temp, MAX_ARG_LEN );

			String[0] = '\0';
			GetUserMacroValue( temp, String );
			return 0;
		}
		else {
			DisplayErrorHeader();
			_wcsupr_s( ch, MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad macro name: \"#%s\"\r\n", ch );
			return -1;
		}
	}

	if ( *ch == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing macro name\r\n" );
		return -1;
	}

	rv = ParseInt( ch, &Number, &ch );

	if ( rv || ( *ch != '\0' ) ) {
		_wcsupr_s( String, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad macro number: \"%s\"\r\n", String );
		return -1;
	}

	String[0] = '\0';
	GetMacroValue( Number, String );
	return 0;
}



int FreeLinkedList( void **FirstItem )
{
	if ( !FirstItem )
		return -1;

	size_t *This		= (size_t*) *FirstItem;
	size_t Next			= NULL;


	while ( This ) {
		//	Printf( L"   * Freeing memory at 0x%08X\r\n", This );
		Next = *This;
		free( This );
		This = (size_t*) Next;
	}

	return 0;
}



int FreeLinkedList( _ListMatchEntry **FirstItem )
{
	return FreeLinkedList( (void**) FirstItem );
}



DLLExports int WINAPI apingmacro( LPTSTR inString )
{
	if ( !inString )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]		= L"";
	int NumSet						= 0;
	int NumDeleted					= 0;
	unsigned int Number				= 0;
	unsigned int Pause				= 0;
	BOOL ShowCount					= TRUE;
	BOOL ListAll					= FALSE;
	BOOL Numeric					= FALSE;
	BOOL NoHighlight				= FALSE;
	wchar_t *ch						= NULL;

	_ListMatchEntry *FirstMatch		= NULL;
	_ListMatchEntry *LastMatch		= NULL;
	_ListMatchEntry *NewMatch		= NULL;

	int j							= 0;
	int rv							= 1;



	CL_NoPausing();

	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( IsOption( arg[i], L"/?", FALSE ) ) {
				ShowCmdHelp( APingMacroHelpText, FALSE );
				rv = 0;
				goto Abort;
			}
			else if ( IsOption( arg[i], L"/??", FALSE ) ) {
				ShowCmdHelp( APingMacroHelpText, TRUE );
				rv = 0;
				goto Abort;
			}
			else if ( IsOption( arg[i], L"/D", FALSE ) )
				NoHighlight = TRUE;
			else if ( IsOption( arg[i], L"/L", FALSE ) )
				ListAll = TRUE;
			else if ( IsOption( arg[i], L"/P", TRUE ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) ) {
					rv = 1;
					goto Abort;
				}
			}
			else if ( ( arg[i][0] == '/' ) || ( arg[i][0] == '-' ) ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown option: \"%s\"\r\n", arg[i] );
				rv = 1;
				goto Abort;
			}
			else {
				ch = arg[i];
				if ( *ch == '#' )
					ch++;

				if ( ( *ch == '\0' ) || IsColonOrEquals( *ch ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing macro name: \"%s\"\r\n", arg[i] );
					rv = 1;
					goto Abort;
				}

				if ( iswdigit( *ch ) ) {			//	numeric macro:

					if ( ParseInt( ch, &Number, &ch ) ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( L"Bad macro number: \"%s\"\r\n", arg[i] );
						goto Abort;
					}

					temp[0] = '\0';
					Numeric = TRUE;
				}
				else {								//	string-named macro:

					j = 0;
					temp[j] = '\0';
					Number = 0;

					while ( *ch && !IsColonOrEquals( *ch ) ) {
						if ( j < ( SHORT_BUF_LEN - 1 ) ) {
							temp[j++] = towupper( *ch );
							temp[j] = '\0';
						}

						ch++;
					}

					Numeric = FALSE;
				}

				if ( *ch == '\0' ) {					//	add match entry to list:

					NewMatch = (_ListMatchEntry*) malloc( sizeof( _ListMatchEntry ) + ( wcslen( temp ) * sizeof( wchar_t ) ) );

					if ( !NewMatch ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Error allocating match entry\r\n" );
						rv = 2;
						goto Abort;
					}

					if ( Numeric ) {
						NewMatch->MacroName[0] = '\0';
						NewMatch->Number = Number;
					}
					else {
						wcscpy_s( NewMatch->MacroName, wcslen( temp ) + 1, temp );
						NewMatch->Number = 0;
					}

					NewMatch->Next = NULL;

					if ( FirstMatch ) {
						LastMatch->Next = NewMatch;
						LastMatch = NewMatch;
					}
					else
						FirstMatch = LastMatch = NewMatch;

				}
				else if ( IsColonOrEquals( *ch ) ) {	//	define or remove a macro:
					ch++;

					if ( Numeric )
						SetMacroValue( Number, ch );
					else {
						if ( !IsValidUserMacroName( temp ) ) {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Bad macro name: \"#%s\"\r\n", temp );
							rv = 1;
							goto Abort;
						}

						if ( wcslen( ch ) >= NI_MAXHOST ) {
							_wcsupr_s( ch, MAX_ARG_LEN );
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Hostname too long: \"%s\"\r\n", ch );
							rv = 1;
							goto Abort;
						}

						SetMacroValue( temp, ch );
					}

					if ( *ch )
						NumSet++;
					else
						NumDeleted++;
				}
				else {									//	anything else is bad
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					rv = 1;
					goto Abort;
				}
			}

		}
	}


	CL_PickHighlightColor( NoHighlight );

	if ( Pause )
		CL_SetPauseLines( Pause );


	if ( !NumSet && !NumDeleted && !FirstMatch )
		ListAll = TRUE;

	if ( ListAll ) {
		if ( ListMacros( ShowCount, NULL ) == USER_ABORT )
			return 3;
	}
	else if ( FirstMatch ) {
		if ( ListMacros( ShowCount, FirstMatch ) == USER_ABORT )
			return 3;
	}

	FreeLinkedList( &FirstMatch );
	return 0;


Abort:

	FreeLinkedList( &FirstMatch );
	return rv;
}



int PingTime2( LPTSTR Hostname, unsigned int Timeout, unsigned int PacketSize, BYTE Ttl, unsigned int Flags )
{
	if ( !Hostname )
		return -3;
	else if ( Hostname[0] == '\0' )
		return -3;


	wchar_t temp[SHORT_BUF_LEN]			= L"";

	HANDLE IcmpHandle					= NULL;
	SOCKADDR_STORAGE HostAddress;
	sockaddr_in *HostAddressV4			= (sockaddr_in*) &HostAddress;
	sockaddr_in6 *AddressV6				= (sockaddr_in6*) &HostAddress;

	int Type							= 0;
	DWORD AddressV4						= 0;

	sockaddr_in6 SourceAddressV6;
	BYTE *RequestData					= NULL;
	IP_OPTION_INFORMATION Options;

	ICMP_ECHO_REPLY *ReplyBuffer4		= NULL;
	ICMPV6_ECHO_REPLY *ReplyBuffer6		= NULL;
	unsigned int ReplyBufferSize		= 0;

	BYTE TryFirst						= AF_INET;
	BYTE TrySecond						= AF_INET6;

	int rv								= -1;



	if ( Flags & PING_FLAG_TRY_V6_FIRST ) {
		TryFirst	= AF_INET6;
		TrySecond	= AF_INET;
	}


	memset( &HostAddress, 0x00, sizeof( SOCKADDR_STORAGE ) );

	if ( LookupHostIP( Hostname, &HostAddress, TryFirst, TrySecond, NULL, NULL ) )
		return -3;

	Type = HostAddress.ss_family;

	if ( Type == AF_INET ) {
		AddressV4 = HostAddressV4->sin_addr.S_un.S_addr;

		if ( Flags & PING_FLAG_DEBUG )
			wprintf( L"   * IPv4 address:  %u.%u.%u.%u\n", ( AddressV4 & 0xff ), ( AddressV4 >> 8 ) & 0xff, ( AddressV4 >> 16 ) & 0xff, ( AddressV4 >> 24 ) );
	}
	else if ( Type == AF_INET6 ) {
		if ( Flags & PING_FLAG_DEBUG ) {
			MakeIPv6String( AddressV6, temp );
			wprintf( L"   * IPv6 address:  %s\n", temp );
		}
	}
	else
		return -3;


	if ( Timeout < ( PING_MIN_TIMEOUT * 1000 ) )
		Timeout = PING_MIN_TIMEOUT * 1000;
	else if ( Timeout > ( PING_MAX_TIMEOUT * 1000 ) )
		Timeout = PING_MAX_TIMEOUT * 1000;

	if ( ( Ttl < TTL_MIN ) || ( Ttl > TTL_MAX ) )
		Ttl = TTL_DEFAULT;

	memset( &Options, 0x00, sizeof( IP_OPTION_INFORMATION ) );
	Options.Ttl = Ttl;



	RequestData = (BYTE*) malloc( PacketSize );
	if ( !RequestData )
		return -4;

	for ( unsigned int i = 0; i < PacketSize; i++ ) {
		unsigned int Value;
		rand_s( &Value );
		RequestData[i] = Value & 0xff;
	}


	if ( Type == AF_INET ) {
		ReplyBufferSize = sizeof( ICMP_ECHO_REPLY ) + PacketSize + 16;
		ReplyBuffer4 = (ICMP_ECHO_REPLY*) malloc( ReplyBufferSize );
		if ( !ReplyBuffer4 ) {
			rv = -4;
			goto Abort;
		}

		memset( ReplyBuffer4, 0x00, ReplyBufferSize );

		IcmpHandle = IcmpCreateFile();
		if ( !IcmpHandle || ( IcmpHandle == INVALID_HANDLE_VALUE ) ) {
			rv = -4;
			goto Abort;
		}

		if ( IcmpSendEcho( IcmpHandle, AddressV4, RequestData, PacketSize, &Options, ReplyBuffer4, ReplyBufferSize, Timeout ) ) {
			if ( ( ReplyBuffer4->Status == IP_SUCCESS ) && ( ReplyBuffer4->Address == AddressV4 ) )
				rv = ReplyBuffer4->RoundTripTime;
			else if ( ( ReplyBuffer4->Status == IP_DEST_NET_UNREACHABLE ) || ( ReplyBuffer4->Status == IP_DEST_HOST_UNREACHABLE ) || ( ReplyBuffer4->Status == IP_DEST_PROT_UNREACHABLE ) || ( ReplyBuffer4->Status == IP_DEST_PORT_UNREACHABLE ) )
				rv = -2;
			else if ( ReplyBuffer4->Status == IP_REQ_TIMED_OUT )
				rv = -1;

			if ( ( rv >= 0 ) && ( Flags & PING_FLAG_COMPARE_DATA ) )
				if ( memcmp( RequestData, ReplyBuffer4->Data, PacketSize ) )
					rv = ERR_DATA_MISMATCH;
		}
	}
	else {
		ReplyBufferSize = sizeof( ICMPV6_ECHO_REPLY ) + PacketSize + 16 + sizeof( __IO_STATUS_BLOCK );
		ReplyBuffer6 = (ICMPV6_ECHO_REPLY*) malloc( ReplyBufferSize );
		if ( !ReplyBuffer6 ) {
			rv = -4;
			goto Abort;
		}

		memset( ReplyBuffer6, 0x00, sizeof( ReplyBufferSize ) );

		memset( &SourceAddressV6, 0x00, sizeof( sockaddr_in6 ) );
		SourceAddressV6.sin6_family		= AF_INET6;
		SourceAddressV6.sin6_addr		= in6addr_any;

		IcmpHandle = Icmp6CreateFile();
		if ( !IcmpHandle || ( IcmpHandle == INVALID_HANDLE_VALUE ) ) {
			rv = -4;
			goto Abort;
		}

		if ( Icmp6SendEcho2( IcmpHandle, NULL, NULL, NULL, &SourceAddressV6, AddressV6, RequestData, PacketSize, &Options, ReplyBuffer6, ReplyBufferSize, Timeout ) ) {
			if ( ( ReplyBuffer6->Status == IP_SUCCESS ) && !memcmp( &( ReplyBuffer6->Address.sin6_addr ), &( AddressV6->sin6_addr ), sizeof( IN6_ADDR ) ) )
				rv = ReplyBuffer6->RoundTripTime;
			else if ( ( ReplyBuffer6->Status == IP_DEST_NET_UNREACHABLE ) || ( ReplyBuffer6->Status == IP_DEST_HOST_UNREACHABLE ) || ( ReplyBuffer6->Status == IP_DEST_PROT_UNREACHABLE ) || ( ReplyBuffer6->Status == IP_DEST_PORT_UNREACHABLE ) )
				rv = -2;
			else if ( ReplyBuffer6->Status == IP_REQ_TIMED_OUT )
				rv = -1;

			if ( ( rv >= 0 ) && ( Flags & PING_FLAG_COMPARE_DATA ) )
				if ( memcmp( RequestData, (BYTE*) ReplyBuffer6 + sizeof( ICMPV6_ECHO_REPLY ), PacketSize ) )
					rv = ERR_DATA_MISMATCH;
		}
	}


Abort:

	IcmpCloseHandle( IcmpHandle );
	free( RequestData );
	free( ReplyBuffer4 );
	free( ReplyBuffer6 );
	return rv;
}



DLLExports int WINAPI f_pingtime( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int Timeout		= 60;
	unsigned int PacketSize		= 64;
	unsigned int Ttl			= GetDefaultTTL();
	unsigned int Flags			= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing hostname\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Timeout, NULL ) ) {
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad timeout: \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( ( Timeout < PING_MIN_TIMEOUT ) || ( Timeout > PING_MAX_TIMEOUT ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Timeout must be %u to %u: %u\r\n", PING_MIN_TIMEOUT, PING_MAX_TIMEOUT, Timeout );
			return -1;
		}
	}

	if ( arglen[2] ) {
		if ( ParseInt( arg[2], &PacketSize, NULL ) ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad packetsize: \"%s\"\r\n", arg[2] );
			return -1;
		}

		if ( ( PacketSize < PING_MIN_PKTSIZE ) || ( PacketSize > PING_MAX_PKTSIZE ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Packetsize must be %u to %u: %u\r\n", PING_MIN_PKTSIZE, PING_MAX_PKTSIZE, PacketSize );
			return -1;
		}
	}

	if ( arglen[3] ) {
		if ( ParseInt( arg[3], &Ttl, NULL ) ) {
			_wcsupr_s( arg[3], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad TTL: \"%s\"\r\n", arg[3] );
			return -1;
		}

		if ( ( Ttl < TTL_MIN ) || ( Ttl > TTL_MAX ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"TTL must be %u to %u: %u\r\n", TTL_MIN, TTL_MAX, Ttl );
			return -1;
		}
	}

	if ( arglen[4] ) {
		if ( ParseInt( arg[4], &Flags, NULL ) ) {
			_wcsupr_s( arg[4], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags: \"%s\"\r\n", arg[4] );
			return -1;
		}
	}

	wsprintf( String, L"%d", PingTime2( arg[0], Timeout * 1000, PacketSize, Ttl, Flags ) );
	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/aping.html";
	piInfo.pszDescription	= (LPTSTR) L"Adds an audible ping command to TCC.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	memset( &WsaData, 0x00, sizeof( WSADATA ) );
	if ( WSAStartup( 0x0101, &WsaData ) ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unable to initialize WinSock API!\r\n" );
		return 1;
	}

	if ( CheckMidiDisabled() )
		MidiHandle = NULL;
	else {
		MidiHandle = OpenSharedMidi();		//	midiOutOpen( &MidiHandle, 0, NULL, NULL, CALLBACK_NULL );

		if ( !MidiHandle ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ),L"Unable to open MIDI handle!\r\n" );
			WSACleanup();
			return 1;
		}
	}

	memset( &OSVerInfo, 0x00, sizeof( OSVERSIONINFO ) );
	OSVerInfo.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
	GetVersionEx( &OSVerInfo );


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	CL_Initialize();
	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	WSACleanup();

	if ( MidiHandle )
		CloseSharedMidi( MidiHandle );		//	midiOutClose( MidiHandle );

#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif

	CL_Shutdown();
	return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"aping" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( APingHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"apinghelp" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( APingHelpHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"apingmacro" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( APingMacroHelpText ) );
		return 1;
	}
	else
		return 0;
}



DLLExports int WINAPI apinghelp( LPTSTR inString )
{
	return PluginHelpCmd( inString, APingHelpHelpText );
}



/*

	0.76.3		2021-02-11
				Changes to allow APing to share a MIDI handle with my other plugins. Also,
				changed the filenames:  APing.dll is now the 64-bit build; the 32-bit
				version is APing-x86.

	1.0.0.0		2021-10-20
				Tweaks to the way help text is displayed.  Documented highlighting, and
				added a /D option to both APING and APINGMACRO.

	1.0.1.0		2022-12-15
				Improved LookupHostIP() :  If GetAddrInfoW() doesn't pick up an IPv6
				address, try DnsQuery_W().

	1.0.1.1		2023-06-13
				Updated the help system to work around an issue in TCC v20.00.30.

	1.0.2.0		2023-07-10
				Setting the volume with APING /Y automatically unmutes audio.  Updated to
				the current versions of ParseArgs.cpp, SoundVol.cpp, and conlist.cpp.

	1.0.2.1		2023-08-29

	1.0.2.2		2023-10-12
				Updated the plugin's web address.

	1.0.3.0		2024-01-04
				Updated to conlist.cpp v1.1.

*/
